#define NeedFunctionPrototypes 1
#include <X11/Xlib.h>
#include <X11/keysym.h>
#if XlibSpecificationRelease != 6
   #error Requires X11R6
#endif

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/input.h>
#include <linux/kd.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>

#define EVDEV          "/dev/input/event"
#define PROGNAME       "actfb2x"
#define VERSION        "0.1"

#define MAX_VALUE      0xff

static char    *displayname    = NULL;
static Display *display        = NULL;
static Window window = 0;
typedef struct keycode_lircmap {
   int value;
   KeySym keycode;
   char * lircString;
} key2lirc;

static key2lirc activyKeys[] = {
   { 0x82, KEY_VOLUMEUP, "vol+"},
   { 0x83, KEY_F2, "green"},   
   { 0x84, KEY_SEARCH, "goto"},
   { 0x85, KEY_EXIT, "exit"},
   { 0x86, KEY_RIGHT, "right"},
   { 0x87, KEY_EJECTCD, "eject"},
   { 0x88, KEY_FASTFORWARD, "ffwd"},
   { 0x8c, KEY_LEFT, "left"},
   { 0x8e, KEY_LEFT, "kleft"},
   { 0x8f, KEY_PLAY, "play"},
   { 0x90, KEY_POWER, "power"},
   { 0x91, KEY_F4, "blue"},
   { 0x93, KEY_LEFT, "left"},
   { 0x94, KEY_DOWN, "down"},
   { 0x95, KEY_NEXTSONG, "jump+"},
   { 0x96, KEY_SCROLLUP, "scr+"},
   { 0x99, KEY_RIGHT, "right"},
   { 0x9e, KEY_KPMINUS, "ch-"},
   { 0x9f, KEY_KPPLUS, "ch+"},
   { 0xa1, KEY_UP, "up"},
   { 0xa2, KEY_PREVIOUSSONG, "jump-"},
   { 0xa3, KEY_RECORD, "rec"},
   { 0xa4, KEY_POWER, "power"},
   { 0xa8, KEY_KPENTER, "ok"},
   { 0xab, KEY_ESC, "clear"},
   { 0xac, KEY_SCROLLDOWN, "scr-"},
   { 0xae, KEY_MENU, "menu"},
   { 0xaf, KEY_HOME, "home"},
   { 0xb0, KEY_MUTE, "mute"},
   { 0xb1, KEY_HELP, "info"},
   { 0xb2, KEY_DOWN, "down"},
   { 0xb9, KEY_F3, "yellow"},
   { 0xbb, KEY_5, "5"},
   { 0xbc, KEY_6, "6"},
   { 0xbd, KEY_4, "4"},
   { 0xbf, KEY_3, "3"},
   { 0xc1, KEY_2, "2"},
   { 0xc2, KEY_0, "0"},
   { 0xc3, KEY_1, "1"},
   { 0xc4, KEY_9, "9"},
   { 0xd5, KEY_KP8, "dup"},
   { 0xd6, KEY_KP2, "ddown"},
   { 0xd8, KEY_7, "7"},
   { 0xd9, KEY_STOP, "stop"},
   { 0xda, KEY_VOLUMEDOWN, "vol-"},
   { 0xe4, KEY_8, "8"},
   { 0xe7, KEY_F1, "red"},
   { 0xe8, KEY_COMPUTER, "keyb"},
   { 0xe9, KEY_KPENTER, "ok"},
   { 0xea, KEY_REWIND, "frew"},
   { 0xed, KEY_UP, "up"},
   { 0xf5, KEY_CANCEL, "cancel"},
   { 0xf6, KEY_CONFIG, "option"},
   { 0xf7, KEY_KP4, "dleft"},
   { 0xf8, KEY_KP6, "dright"},
   { 0xf9, KEY_UNKNOWN, "tv"},
   { 0xfd, KEY_RIGHT, "kright"},
   { 0xfe, KEY_ENTER, "enter"},
   { 0, 0, NULL}
};

static key2lirc keyTable[ MAX_VALUE + 1 ];

/* prototypes */
static int setup_key_table(char * mapFile);
static int setup_event_dev(void);
static void process_key(int efd);
int MyErrorHandler(Display *my_display, XErrorEvent *event);
void SendKeyPressedEvent( KeySym keysym );
void SendEvent(XKeyEvent *event);


static int setup_key_table(char * mapFile)
{
   int i;
   FILE *fd;
   char line[256];

   for( i = 1; i <= MAX_VALUE; i ++ ) {
      keyTable[i].value = keyTable[i].keycode = 0;
      keyTable[i].lircString = NULL;
   }
   keyTable[0].lircString = "pause";

   for( i = 0; activyKeys[i].value > 0 && activyKeys[i].value <= MAX_VALUE; i++ ) {
      keyTable[ activyKeys[i].value ] = activyKeys[i];
      keyTable[ activyKeys[i].value ].keycode = 0;
   }

   fd = fopen(mapFile, "r");
   if( fd == NULL ) {
      fprintf(stderr, "Cannot open <%s>\n", mapFile);
      return 1;
   }
   while( !feof(fd) ) {
      fgets(line, sizeof(line)-1, fd);
      if(line[0] != '#') {
         char *p = strchr( line, '=' );
         if(p && (p > line) && (*(p+1) != '\0')) {
            int found = 0;
            *p = '\0';
            p++;
            *(p+strlen(p)-1)='\0';

            for( i = 1; i <= MAX_VALUE; i ++ ) {
               if( keyTable[i].lircString != 0 && !strcmp(keyTable[i].lircString, line) ) {
                  keyTable[i].keycode = XStringToKeysym(p);
                  if(keyTable[i].keycode == NoSymbol) {
                     fprintf(stderr, "Unknown XKey <%s>\n", p);
                  }
                  found = 1;
                  break;
               }
            }
            if( !found )
               fprintf(stderr, "Unknown Key <%s>\n", line);
         }
      }
   }
   return 0;
}

static int setup_event_dev()
{
   int rfd = -1;
   int i;
   char name[256];
   char info[128];

   for( i = 0; i < 16; i++ ) {
      sprintf( name, "%s%d", EVDEV, i );
      if( (rfd = open (name, O_RDONLY)) >= 0 ) {
         ioctl (rfd, EVIOCGNAME (sizeof (info)), info);
         if( strstr( info, "keyboard" ) ) {
            printf("Using: %s (%s)\n", name, info);
            break;
         }
         printf("Not using: %s (%s)\n", name, info);
         close( rfd );
         rfd = -1;
      }
      else
         fprintf(stderr,"Can't open input event device %s\n", name);
   }
   return rfd;
}


static void process_key(int efd)
{
   struct input_event event[64];
   int len, idx;

   len = read(efd, event, sizeof(struct input_event) * 64);

   if( len < (int) sizeof(struct input_event) ) {
      fprintf(stderr, "Error processing input event %s\n", strerror(errno));
      return;
   }

   if( event[1].value < 1 || event[0].value > MAX_VALUE )
      return;

   idx = event[0].value;
   if( keyTable[idx].keycode != 0 )
      SendKeyPressedEvent(keyTable[idx].keycode);

   return;
}

int MyErrorHandler(Display *my_display, XErrorEvent *event)
{
   fprintf(stderr, "%s: XSendEvent(0x%lx) failed.\n", PROGNAME, window);
   return 1;
}

void SendEvent(XKeyEvent *event)
{
   XSync(display, False);
   XSetErrorHandler(MyErrorHandler);
   XSendEvent(display, window, True, KeyPressMask, (XEvent*)event);
   XSync(display, False);
   XSetErrorHandler(NULL);
}

void SendKeyPressedEvent( KeySym keysym )
{
   XKeyEvent     event;

   int Junk;

   XGetInputFocus(display, &window, &Junk);

   if( window == (Window)-1 ) {
      window = RootWindow(display, 0); // XXX nonzero screens?
   }
   event.display = display;
   event.window  = window;
   event.root    = RootWindow(display, 0); // XXX nonzero screens?
   event.subwindow  = None;
   event.time    = CurrentTime;
   event.x    = 1;
   event.y    = 1;
   event.x_root  = 1;
   event.y_root  = 1;
   event.same_screen   = True;
   event.type    = KeyPress;
   event.state      = 0;

   //
   // press down shift keys one at a time...
   //

   event.keycode = XKeysymToKeycode(display, keysym);
   // fprintf(stderr, "code: 0x%x, %d\n", event.keycode, event.keycode );
   SendEvent(&event);

   event.type = KeyRelease;
   SendEvent(&event);
}

int main(int argc, char **argv)
{
   fd_set permset, fdset;
   int event_fd;

   printf( "\n%s Version %s\n", PROGNAME, VERSION );
   event_fd = setup_event_dev();
   if( event_fd < 0 ) {
      fprintf(stderr,"No input device found\n" );
      return 1;
   }

   if( setup_key_table( argv[1] ) )
      return 1;

   signal(SIGPIPE, SIG_IGN);

   FD_ZERO(&permset);
   FD_SET(event_fd, &permset);

   if( displayname == NULL )
      displayname = getenv("DISPLAY");

   if( displayname == NULL )
      displayname = ":0.0";

   display = XOpenDisplay(displayname);

   if( display == NULL ) {
      fprintf(stderr, "Can't open display `%s'.\n", displayname);
      exit(1);
   }

   while( 1 ) {
      fdset = permset;

      if( select(event_fd + 1, &fdset, NULL, NULL, NULL) < 0 ) {
         if( errno == EINTR )
            continue;
         fprintf(stderr, "Error during select(): %s\n", strerror(errno));
         return -1;
      }

      if( FD_ISSET(event_fd, &fdset) )
         process_key(event_fd);
   }

   XCloseDisplay(display);
   return 0;
}
